//
// Подключаем библиотеки
//

// Подключаем датчики температуры
#include <OneWire.h>
#include <DallasTemperature.h>

// Датчки подключены к 3 порту
OneWire oneWire(3);
DallasTemperature sensors(&oneWire);

DeviceAddress Thermometer1 = {
  0x28, 0x9C, 0xD3, 0x07, 0x00, 0x00, 0x80, 0xC8
  // адрес датчика DS18B20 289CD307000080C8
  //0x28, 0x9D, 0xC6, 0x07, 0x00, 0x00, 0x80, 0x51
  //Device 0 Address: 28 9D C6 07 00 00 80 51
};   
DeviceAddress Thermometer2 = {
  0x28, 0xBA, 0xC9, 0x07, 0x00, 0x00, 0x80, 0x0D
  // адрес датчика DS18B20 28BAC9070000800D
  //0x28, 0xF7, 0xC7, 0x07, 0x00, 0x00, 0x80, 0x90
  //Device 1 Address: 28 F7 C7 07 00 00 80 90
};   

// Подключаем экран
#include <LiquidCrystal_I2C.h>
LiquidCrystal_I2C lcd(0x27, 16, 2); // Задаем адрес и размерность дисплея.

// Подключаем ИК
#include <IRremote.h>
IRrecv irrecv(2);
decode_results results;

//
// Объявляем кучу глобальных переменных
//

int iv_mode = 1; // Режим работы: 1 - Автомат, 2 - Настройки, 3 - Ручной режим
bool iv_back_light = 1; // Подсветка экрана
unsigned long old_time = 30000;  // Переменная времени
unsigned long old_time_print = 30000;  // Переменная времени для обновления показателей

float temp_1 = 20; // Показания первого датчика (наружная)
float temp_2 = 20; // Показания второго датчика (внутренная)
float temp_v = 0; // Температура контура отопления
float iv_k = 1.2; // Кривая отопления
float temp_u = 21; // Пользовательские настройки температуры
float temp_n = 0; // Температура контура отопления в зависимости от наружной температуры
float temp_k = 0; // Поправка на желаемую комнатную температуру
float temp_t = 0; // Поправка термостата
float delta_k = 3; // Шаг поправки комнатной температуры
float delta_t = 3; // Шаг поправки термостата
int int_out = 0; // Выходное значение платы управляения

// Текущая настройка
int current_settings = 1;
  // Для режима настроек
  // 1 - Кривая отопления
  // 2 - Шаг поправки комнатной температуры
  // 3 - Шаг поправки термостата

  // Для автоматического режима 
  // 1 - Основной экран
  // 2 - Экран контура отопления
  // 3 - Экран статистики

// Статистика
float st_temp_1_max = -50;
float st_temp_1_min = 50;
int st_int_out_max = 0;
int st_int_out_min = 90;


// Температура котла зависит от значений платы управления нелинейно, поэтому воспользуемся табличными данными
// Номер элемента в массиве совпадает с расчетным значением int_out
//int vaillant[] = {0, 0, 32, 35, 38, 41, 43, 46, 48, 51, 53, 55, 57, 59, 60, 62, 63, 64, 66, 68, 69, 70, 71, 72, 74, 75, 76, 77, 78, 79, 79, 80};
int vaillant[] = {24, 28, 32, 35, 38, 41, 43, 46, 48, 51, 53, 55, 57, 59, 60, 62, 63, 64, 66, 68, 69, 70, 71, 72, 74, 75, 76, 77, 78, 79, 79, 80};


void setup() {
  //
  // Инициализируем все устройства
  //

  lcd.init();  // Инициализация lcd
  lcd.backlight();  // Включаем подсветку
  lcd.setCursor(0, 0);
  lcd.print("Hello!");

  // Запускаем порт обмена данными по USB
  Serial.begin(9600);

  // Запуск температурных датчиков
  sensors.begin();
  sensors.setResolution(Thermometer1, 12);
  sensors.setResolution(Thermometer2, 12);

  // ИК пульт
  irrecv.enableIRIn();

  // Активируем реле 5-9
  lcd.setCursor(0, 1);
  for (int i = 5; i <= 9; i++) {
    pinMode (i, OUTPUT);
    digitalWrite (i, HIGH);
  }

  // Включаем предохранитель
  pinMode (4, OUTPUT);
  digitalWrite (4, LOW);
}

void loop() {
  //
  // Основной цикл
  //

  unsigned long time_abs = 0; // Разница во времени
  unsigned long time_abs_print = 0; // Разница во времени

  // сколько времени прошло
  time_abs = millis() - old_time;
  time_abs = abs(time_abs);

  time_abs_print = millis() - old_time_print;
  time_abs_print = abs(time_abs_print);

  // Автоматический режим
  if (iv_mode == 1) {   
    // 5 минут паузы 5*60*1000
    if (time_abs > 300000) {
      old_time = millis();
      get_temp(); // Получем температуру
      calc_temp(); // Расчитываем температуру контура отопления
      calc_int_out(); // Расчитываем коэффциент платы управления
      set_relay(); // Выставляем реле
      stat(); // Сохраняем для статистики
      lcd_print(); // Выводим результаты на дисплей

      // !!! Передача текущих значений на компьютер
      // желательно каждую минуту
      // serial_print();
      
    }

    // если прошла минута
    if (time_abs_print > 60000){
      old_time_print = millis();
      get_temp(); // Получем температуру
      stat(); // Сохраняем для статистики
      lcd_print(); // Выводим результаты на дисплей
      com_print(); // Посылаем на комп
     }
  }

  // Ручной режим
  if (iv_mode == 3) {
    // Сработал счетчик времени
    // Обновляем показания температуры каждые 5 секунд
    if (time_abs > 5000) {
      old_time = millis();
      get_temp(); // Получем температуру
      lcd_print(); // Выводим результаты на дисплей
    }   
  }
  ir_key();  //проверяем кнопочки



}

void ir_key() {
  //
  // Обрабатываем нажатие кнопок в разных режимах
  //

  if (irrecv.decode(&results)) {
    //
    // Обработка нажатий в общем случае
    //
    switch (results.value) {
      case 0xFF52AD:
        // #
        // переключаем подсветку
        if (iv_back_light) {
          iv_back_light = 0;
          lcd.noBacklight();
        } else {
          iv_back_light = 1;
          lcd.backlight();
        }
        break;
      
      case 0xFF42BD:
      // *
        // Сброс статистики
        st_temp_1_max = -50;
        st_temp_1_min = 50;
        st_int_out_max = 0;
        st_int_out_min = 90;
        stat();
        lcd_print(); // Выводим результаты на дисплей
        break;

      // управление режимами
      // 7 - автоматический режим
      case 0xFF10EF:
        iv_mode = 1;
        get_temp(); // Получем температуру
        calc_temp(); // Расчитываем температуру контура отопления
        calc_int_out(); // Расчитываем коэффциент платы управления
        set_relay(); // Выставляем реле
        lcd_print(); // Выводим результаты на дисплей
        break;

      // 8 - режим настроек
      case 0xFF38C7:
        iv_mode = 2;
        get_temp(); // Получем температуру
        lcd_print(); // Выводим результаты на дисплей
        break;

      // 9 - ручной режим
      case 0xFF5AA5:
        iv_mode = 3;
        set_relay();
        get_temp(); // Получем температуру
        lcd_print(); // Выводим результаты на дисплей
        break;

    }

    //
    // Автоматический режим
    //
    if (iv_mode == 1) {
      switch (results.value) {
        // вверх
        case 0xFF629D:
          // увеличиваем желаемую температруру на полградуса
          temp_u = temp_u + 0.5;
          if (temp_u > 27) temp_u = 27;
          lcd_print(); // Выводим результаты на дисплей
          break;

        // вниз
        case 0xFFA857:
          // уменьшаем желаемую температруру на полградуса
          temp_u = temp_u - 0.5;
          if (temp_u < 15) temp_u = 15;
          lcd_print(); // Выводим результаты на дисплей
          break;

        // вправо/влево - переход между экранами
        // вправо 
        case 0xFFC23D:
          current_settings++;
          if (current_settings > 3) current_settings = 1;
          lcd_print(); // Выводим результаты на дисплей
          break;
        // влево
        case 0xFF22DD:
          current_settings--;
          if (current_settings < 1) current_settings = 3;
          lcd_print(); // Выводим результаты на дисплей
          break;
      }
    }

    //
    // Режим настроек
    //
    if (iv_mode == 2) {
      switch (results.value) {
        // вправо/влево - переход между настройками
        // вправо 
        case 0xFFC23D:
          current_settings++;
          if (current_settings > 3) current_settings = 1;
          lcd_print(); // Выводим результаты на дисплей
          break;
        // влево
        case 0xFF22DD:
          current_settings--;
          if (current_settings < 1) current_settings = 3;
          lcd_print(); // Выводим результаты на дисплей
          break;
          
        // вверх/вниз - меняем настройку
        // вверх
        case 0xFF629D:
          // 1 - Кривая отопления
          if (current_settings == 1) {
            iv_k = iv_k + 0.05;
            lcd_print(); // Выводим результаты на дисплей
          }
          // 2 - Шаг поправки комнатной температуры          
          if (current_settings == 2) {
            delta_k = delta_k + 0.5;
            lcd_print(); // Выводим результаты на дисплей
          }
          // 3 - Шаг поправки термостата
          if (current_settings == 3) {
            delta_t = delta_t + 0.1;
            lcd_print(); // Выводим результаты на дисплей
          }
          break;          

        // вниз
        case 0xFFA857:
          // 1 - Кривая отопления
          if (current_settings == 1) {
            iv_k = iv_k - 0.05;
            lcd_print(); // Выводим результаты на дисплей
          }
          // 2 - Шаг поправки комнатной температуры          
          if (current_settings == 2) {
            delta_k = delta_k - 0.5;
            lcd_print(); // Выводим результаты на дисплей
          }
          // 3 - Шаг поправки термостата
          if (current_settings == 3) {
            delta_t = delta_t - 0.1;
            lcd_print(); // Выводим результаты на дисплей
          }
          break;          

        // В режиме настроек можно играть с температурой, для проверки выставленных значений
        // 1-4 управление внешеней температурой
        // 2-5 управление внутренней температурой
        // 1
        case 0xFF6897:
          temp_1 = temp_1+1;
          lcd_print(); // Выводим результаты на дисплей
          break;
        // 4
        case 0xFF30CF:
          temp_1 = temp_1-1;
          lcd_print(); // Выводим результаты на дисплей
          break;
        // 2
        case 0xFF9867:
          temp_2 = temp_2+1;
          lcd_print(); // Выводим результаты на дисплей
          break;
        // 5
        case 0xFF18E7:
          temp_2 = temp_2-1;
          lcd_print(); // Выводим результаты на дисплей
          break;

      }          
        
    }

    //
    // Ручной режим
    //
    if (iv_mode == 3) {
      switch (results.value) {
        // вверх
        case 0xFF629D:
          // увеличиваем вывод;
          int_out++;
          if (int_out > 31) int_out = 0;
          set_relay();
          lcd_print();
          break;
        case 0xFFA857:
          // вниз
          // уменьшаем вывод;
          int_out--;
          if (int_out < 0) int_out = 31;
          set_relay();
          lcd_print();
          break;
      }
    }

    // Сбрасываем кнопки
    irrecv.resume();
  }
}

void lcd_print() {
  //
  // Функция выводит текущую информацию на экран
  //

  // Автоматический режим
  if (iv_mode == 1) {
    // 1 - Основной экран
    if (current_settings == 1) {
      lcd.setCursor(0, 0);
      lcd.print("                ");
      lcd.setCursor(0, 0);
      lcd.print("AM: ");
      // Выводим желаемую температру
      lcd.print(temp_u);
      // Выводим расчетное значение контура отопления
      lcd.print(" (");
      lcd.print(vaillant[int_out]);
      lcd.print(")");

      // Выводим значение температурных датчиков
      lcd.setCursor(0, 1);
      lcd.print("                ");
      lcd.setCursor(0, 1);
      //lcd.print("T1:");
      lcd.print(temp_1);
      lcd.setCursor(8, 1);
      //lcd.print("T2:");
      lcd.print(temp_2);
    }

    // 2 - Экран контура отопления
    if (current_settings == 2) {
      lcd.setCursor(0, 0);
      lcd.print("                ");
      lcd.setCursor(0, 0);
      // temp_v = temp_n + temp_k + temp_t;
      lcd.print("V ");
      lcd.print(temp_v);
      lcd.print(" N ");
      lcd.print(temp_n);

      lcd.setCursor(0, 1);
      lcd.print("                ");
      lcd.setCursor(0, 1);
      lcd.print("K ");
      lcd.print(temp_k);
      lcd.print(" T ");
      lcd.print(temp_t);
    }
  
    // 3 - Экран статистики
    if (current_settings == 3) {
      lcd.setCursor(0, 0);
      lcd.print("                ");
      lcd.setCursor(0, 0);
      lcd.print(st_temp_1_min);
      lcd.setCursor(8, 0);
      lcd.print(st_temp_1_max);
      lcd.setCursor(0, 1);
      lcd.print("                ");
      lcd.setCursor(0, 1);
      lcd.print(vaillant[st_int_out_min]);
      lcd.setCursor(8, 1);
      lcd.print(vaillant[st_int_out_max]);
    }

  }

  // Режим настройки
  if (iv_mode == 2) {

    // 1 - Кривая отопления
    if (current_settings == 1) {
      lcd.setCursor(0, 0);
      lcd.print("                ");
      lcd.setCursor(0, 0);
      lcd.print("SE: K = ");
      lcd.print(iv_k);
      // Выводим значение температурных датчиков
      lcd.setCursor(0, 1);
      lcd.print("                ");
      lcd.setCursor(0, 1);
      lcd.print("T1:");
      lcd.print(temp_1);

      // Выводим расчетное значение контура отопления в зависимости от кривой
      calc_temp();
      lcd.setCursor(8, 1);
      lcd.print(" Tn:");
      lcd.print(temp_n);
    }

    // 2 - Шаг поправки комнатной температуры
    if (current_settings == 2) {
      lcd.setCursor(0, 0);
      lcd.print("                ");
      lcd.setCursor(0, 0);
      lcd.print("SE: DeltaK = ");
      lcd.print(delta_k);
      // Выводим значение температурных датчиков
      lcd.setCursor(0, 1);
      lcd.print("                ");
      lcd.setCursor(0, 1);
      lcd.print("T2:");
      lcd.print(temp_2);
      // Выводим расчетное значение поправки
      calc_temp();
      lcd.setCursor(8, 1);
      lcd.print(" Tk:");
      lcd.print(temp_k);
    }
    
    // 3 - Шаг поправки термостата
    if (current_settings == 3) {
      lcd.setCursor(0, 0);
      lcd.print("                ");
      lcd.setCursor(0, 0);
      lcd.print("SE: DeltaT = ");
      lcd.print(delta_t);
      // Выводим значение температурных датчиков
      lcd.setCursor(0, 1);
      lcd.print("                ");
      lcd.setCursor(0, 1);
      lcd.print("T2:");
      lcd.print(temp_2);
      // Выводим расчетное значение поправки
      calc_temp();
      lcd.setCursor(8, 1);
      lcd.print(" Tt:");
      lcd.print(temp_t);
    
    }

  }

  // Ручной режим
  if (iv_mode == 3) {
    // Функция выводит текущее значение реле на экран
    lcd.setCursor(0, 0);
    lcd.print("                ");
    lcd.setCursor(0, 0);
    if (int_out < 10) lcd.print("0");
    lcd.print(int_out);
    lcd.print(": ");
    for (int j = 4; j >= 0; j--) {
      bool bcl_bit;
      // читаем текущий бит
      bcl_bit = bitRead(int_out, j);
      lcd.print(bcl_bit);
    }
    lcd.print(" (");
    lcd.print(vaillant[int_out]);
    lcd.print(")");
    // Выводим значение температурных датчиков
    lcd.setCursor(0, 1);
    lcd.print("                ");
    lcd.setCursor(0, 1);
    //lcd.print("T1:");
    lcd.print(temp_1);
    lcd.setCursor(8, 1);
    //lcd.print("T2:");
    lcd.print(temp_2);
  }

}

void get_temp() {
  //
  // Функция считывает температуру с датчиков
  //
  float iv_temp = 0;

  // Запрашиваем температуру
  sensors.requestTemperatures();

  // Сохраняем температуру
  iv_temp = sensors.getTempC(Thermometer1);
  // проверка на ошибку
  if (iv_temp > -100) temp_1 = iv_temp;
  iv_temp = sensors.getTempC(Thermometer2);
  if (iv_temp > -100) temp_2 = iv_temp;
  //temp_2 = sensors.getTempC(Thermometer2);
}

void set_relay() {
  //
  // Функция выставлет реле на основе текущего значения вывода платы управления
  //

  for (int j = 4; j >= 0; j--) {
    bool bcl_bit;
    // читаем текущий бит
    bcl_bit = bitRead(int_out, j);
    // выставляем реле
    digitalWrite (j + 5, !bcl_bit);
  }
}

void calc_temp() {
  //
  // Функция расчитывает температуру контура отпления
  //
  float a, b, c, x, t;

  // Расчет поправки термостата
  // Tt = (Tu — T2) × 5
  t = temp_u - temp_2;
  // ограничиваем влияние термостата
  if ( t > 2) {
    t = 2;
  }
  if (t < -2) {
    t = -2;
  }
  temp_t = t * delta_t;

  // Поправка на желаемую комнатную температуру
  // Tk = (Tu — 20) × 5
  temp_k = (temp_u - 20) * delta_k;

  // Температура контура отопления в зависимости от наружной температуры
  // Tn = ax2 + bx + c
  // a = -0,21k — 0,06
  // b = 6,04k + 1,98
  // с = -5,06k + 18,06
  // x = -0.2*t1 + 5

  x = (-0.2 * temp_1) + 5;
  a = (-0.21 * iv_k) - 0.06;
  b = (6.04 * iv_k) + 1.98;
  c = (-5.06 * iv_k) + 18.06;  

  temp_n = (a * x * x) + (b * x) + c;

  // Расчетная температура конура отопления
  // T = Tn + Tk + Tt
  temp_v = temp_n + temp_k + temp_t;

  // Ограничиваем температуру
  if (temp_v > 75) temp_v = 75;
  if (temp_v < 0) temp_v = 0;
}

void calc_int_out() {
  //
  // Расчитываем значение переменной int_out на основе temp_v
  // или выходное значение платы управления на основе расчетной температуры конура отопления
  //

  for (int i = 1; i <= 31; i++) {
    // ищем подходящую темепературу
    if (temp_v <= vaillant[i]) {
      float j1, j2;
      j1 = vaillant[i] - temp_v;
      j2 = temp_v - vaillant[i - 1];
      // выбираем лучшее значение
      if (j2 > j1) {
        int_out = i;
      } else {
        int_out = i - 1;
      }
      break;
    }
  }
}

void com_print() {
  //
  // Выводим по сериал порту текущие значения переменных
  //
  
  // показания датчиков
  Serial.print(temp_1);
  Serial.print("|");
  Serial.print(temp_2);
  Serial.print("|");
  // пользовательские установки
  Serial.print(temp_u);
  Serial.print("|");
  // расчетные значения
  Serial.print(temp_n);
  Serial.print("|");
  Serial.print(temp_k);
  Serial.print("|");
  Serial.print(temp_t);
  Serial.print("|");
  // управление котлом
  Serial.print(int_out);
  Serial.print("|");
  Serial.println(iv_k);
   
}

void stat() {
  if (temp_1 > st_temp_1_max) st_temp_1_max = temp_1;
  if (temp_1 < st_temp_1_min) st_temp_1_min = temp_1;
  if (int_out > st_int_out_max) st_int_out_max = int_out;
  if (int_out < st_int_out_min) st_int_out_min = int_out;
}

